#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2022 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import torch.nn as nn

from tartrl.modules.utils.valuenorm import ValueNorm

class BasePolicyNetwork(nn.Module):
    def __init__(self, args, device):
        super(BasePolicyNetwork, self).__init__()
        self.device = device
        self._use_valuenorm = args.use_valuenorm
        self._use_policy_vhead = args.use_policy_vhead

        if self._use_valuenorm:
            if self._use_policy_vhead:
                self.policy_value_normalizer = ValueNorm(1, device = self.device)
        else:
            if self._use_policy_vhead:
                self.policy_value_normalizer = None

